/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _FLEXIO_DEV_DEBUG_H_
#define _FLEXIO_DEV_DEBUG_H_

#include <libflexio-dev/flexio_dev.h>

extern struct spinlock_s sim_print_lock;

#define TRACE(text, value) do {                        \
	spin_lock(&sim_print_lock);                    \
	__print_sim_str_unlocked(__func__, 0);         \
	__print_sim_str_unlocked(":", 0);              \
	__print_sim_int_unlocked(__LINE__, 0);         \
	__print_sim_str_unlocked(" ", 0);              \
	__print_sim_str_unlocked(text, 20);            \
	__print_sim_hex_unlocked((uint64_t)value, 20); \
	__print_sim_str_unlocked("\n", 0);             \
	spin_unlock(&sim_print_lock);                  \
} while (0)

#define TRACEVAL(val) TRACE(#val, val)

#ifdef CHECK_DEV_INPUT

#define ERRVAL(val) do {                               \
	spin_lock(&sim_print_lock);                    \
	__print_sim_str_unlocked("Param Error - ", 0); \
	__print_sim_str_unlocked(__func__, 0);         \
	__print_sim_str_unlocked(":", 0);              \
	__print_sim_int_unlocked(__LINE__, 0);         \
	__print_sim_str_unlocked(" ", 0);              \
	__print_sim_str_unlocked(#val, 0);             \
	__print_sim_str_unlocked(" is NULL", 0);       \
	__print_sim_str_unlocked("\n", 0);             \
	spin_unlock(&sim_print_lock);                  \
} while (0)


#define RETURN_IF_VALUE_IS_NULL(val) do { \
	if (!(val)) {                     \
		ERRVAL(val);              \
		return;                   \
	}                                 \
} while (0)

#define RETURN_FAIL_IF_VALUE_IS_NULL(val, retval) do { \
	if (!(val)) {                                  \
		ERRVAL(val);                           \
		return retval;                         \
	}                                              \
} while (0)

#define RETURN_ERROR_IF_VALUE_IS_NULL(val) \
	RETURN_FAIL_IF_VALUE_IS_NULL(val, FLEXIO_DEV_STATUS_FAILED)

#define RETURN_NULL_IF_VALUE_IS_NULL(val) \
	RETURN_FAIL_IF_VALUE_IS_NULL(val, NULL)

#else
#define RETURN_IF_VALUE_IS_NULL(val) do {} while (0)

#define RETURN_ERROR_IF_VALUE_IS_NULL(val) do {} while (0)

#define RETURN_NULL_IF_VALUE_IS_NULL(val) do {} while (0)

#define RETURN_FAIL_IF_VALUE_IS_NULL(val, retval) do {} while (0)

#endif

/**
 * @brief Print a string to SimX log.
 *
 * This function prints a given string to SimX log according to user
 * apecified length.
 *
 * @param[in]  str - A pointer to string buffer.
 * @param[in]  len - The length of the string.
 *
 * @return void.
 */
void print_sim_str(const char *str, int len);

/**
 * @brief Print a value in hexadecimal presentation to SimX log.
 *
 * This function prints a given unsigned 64 bit integer value to SimX log
 * in hexadecimal presentation up to a length specified by user.
 * Note that MSB part is dropped if exceeds length.
 *
 * @param[in]  val - A value to be printed.
 * @param[in]  len - The maximal hexadecimal string length to print.
 *
 * @return void.
 */
void print_sim_hex(uint64_t val, int len);

/**
 * @brief Print a value in decimal presentation to SimX log.
 *
 * This function prints a signed given integer value to SimX log
 * in decimal presentation up to a length specified by user.
 * Note that MSB part is dropped if exceeds length.
 *
 * @param[in]  val - A value to be printed.
 * @param[in]  len - The maximal decimal string length to print.
 *
 * @return void.
 */
void print_sim_int(int val, int len);

/**
 * @brief Print a a character to SimX log.
 *
 * This function prints a given integer value as a character
 * to SimX log.
 *
 * @param[in]  val - A value to be printed.
 *
 * @return void.
 */
void print_sim_putc(int val);

/**
 * @brief Service functions. For usage from TRACE macro only
 */
void __print_sim_str_unlocked(const char *str, int len);
void __print_sim_int_unlocked(int val, int len);
void __print_sim_hex_unlocked(uint64_t val, int len);

#endif
