/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file flexio_dev_queue_access.h
 * @page Flex IO SDK dev queue access
 * @defgroup FlexioSDKDevQueueAccess DevQueueAccess
 * @ingroup FlexioSDK
 * Flex IO SDK device API for DPA programs queue access.
 * Provides an API for handling networking queues (WQs/CQs).
 *
 * @{
 */

#ifndef _FLEXIO_DEV_QUEUE_ACCESS_H_
#define _FLEXIO_DEV_QUEUE_ACCESS_H_

#include <libflexio-dev/flexio_dev.h>
#include <libflexio-dev/flexio_dev_queue_types.h>
#include <stdint.h>

/** Flex IO dev WQE control segment types. */
typedef enum {
	FLEXIO_CTRL_SEG_SEND_EN = 0,
	FLEXIO_CTRL_SEG_SEND_RC = 1,
	FLEXIO_CTRL_SEG_LDMA = 2,
	FLEXIO_CTRL_SEG_RDMA_WRITE = 3,
	FLEXIO_CTRL_SEG_RDMA_READ = 4,
	FLEXIO_CTRL_SEG_ATOMIC_COMPARE_AND_SWAP = 5,
	FLEXIO_CTRL_SEG_LSO = 6,
	FLEXIO_CTRL_SEG_NOP = 7,
	FLEXIO_CTRL_SEG_RDMA_WRITE_IMM = 8,
} flexio_ctrl_seg_t;

/**
 * @brief QP/SQ ring doorbell function
 *
 * Rings the doorbell of a QP or SQ in order to alert the HW of pending work.
 *
 * @param[in] dtctx - A pointer to a pointer of flexio_dev_thread_ctx structure.
 * @param[in] pi - Current queue producer index.
 * @param[in] qnum - Number of the queue to update.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_qp_sq_ring_db(struct flexio_dev_thread_ctx *dtctx, uint16_t pi,
					     uint32_t qnum);

/**
 * @brief arm the emulation context
 *
 * @param[in] dtctx - A pointer to a pointer of flexio_dev_thread_ctx structure.
 * @param[in] cqn - CQ number provided by host.
 * @param[in] emu_ctx_id - Emulation context ID, provided by a call on the host to
 *			   flexio_emu_db_to_cq_ctx_get_id.
 */
flexio_dev_status_t flexio_dev_db_ctx_arm(struct flexio_dev_thread_ctx *dtctx,
					  uint32_t cqn, uint32_t emu_ctx_id);

/**
 * @brief force trigger of emulation context
 *
 * @param[in] cqn - CQ number provided by host.
 * @param[in] emu_ctx_id - Emulation context ID, provided by a call on the host to
 *			   flexio_emu_db_to_cq_ctx_get_id.
 */
flexio_dev_status_t flexio_dev_db_ctx_force_trigger(struct flexio_dev_thread_ctx *dtctx,
						    uint32_t cqn, uint32_t emu_ctx_id);

/**
 * @brief Update an EQ consumer index function
 *
 * Updates the consumer index of an EQ after handling an EQE.
 *
 * @param[in] dtctx - A pointer to a pointer of flexio_dev_thread_ctx structure.
 * @param[in] ci - Current EQ consumer index.
 * @param[in] qnum - Number of the EQ to update.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_eq_update_ci(struct flexio_dev_thread_ctx *dtctx, uint32_t ci,
					    uint32_t qnum);

/**
 * @brief Arm CQ function
 *
 * Moves a CQ to 'armed' state.
 * This means that next CQE created for this CQ will result in an EQE on the relevant EQ.
 *
 * @param[in] dtctx - A pointer to a pointer of flexio_dev_thread_ctx structure.
 * @param[in] ci - Current CQ consumer index.
 * @param[in] qnum - Number of the CQ to arm.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_cq_arm(struct flexio_dev_thread_ctx *dtctx, uint32_t ci,
				      uint32_t qnum);

/**
 * @brief Send msix on the cq linked to the msix eq
 *
 * This function trigger msix on the given cq.
 *
 * @param[in] dtctx - A pointer to a flexio_dev_thread_ctx structure.
 * @param[in] cqn - CQ number to trigger db on. Trigger is done via currently configured outbox,
 *                  this can be changed with outbox config API according to CQ.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_msix_send(struct flexio_dev_thread_ctx *dtctx, uint32_t cqn);

/**
 * @brief Get csum OK field from CQE function
 *
 * Parse a CQE for its csum OK field.
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint8_t - csum_ok field value of the CQE.
 */
uint8_t flexio_dev_cqe_get_csum_ok(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get QP number field from CQE function
 *
 * Parse a CQE for its QP number field.
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint32_t - QP number field value of the CQE.
 */
uint32_t flexio_dev_cqe_get_qpn(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get WQE counter filed from CQE function
 *
 * Parse a CQE for its WQE counter field.
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint16_t - WQE counter field value of the CQE.
 */
uint16_t flexio_dev_cqe_get_wqe_counter(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get byte count field from CQE function
 *
 * Parse a CQE for its byte count field.
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint32_t - Byte count field value of the CQE.
 */
uint32_t flexio_dev_cqe_get_byte_cnt(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get the opcode field from CQE function
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint8_t - Opcode field value of the CQE.
 */
uint8_t flexio_dev_cqe_get_opcode(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get the user index field from CQE function
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint32_t - User index field value of the CQE.
 */
uint32_t flexio_dev_cqe_get_user_index(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get CQ number field from EQE function
 *
 * Parse an EQE for its CQ number field.
 *
 * @param[in] eqe - EQE to parse.
 *
 * @return uint32_t - CQ number field value of the EQE.
 */
uint32_t flexio_dev_eqe_get_cqn(struct flexio_dev_eqe *eqe);

/**
 * @brief Get address field from receive WQE function
 *
 * Parse a receive WQE for its address field.
 *
 * @param[in] rwqe - WQE to parse.
 *
 * @return void* - Address field value of the receive WQE.
 */
void *flexio_dev_rwqe_get_addr(struct flexio_dev_wqe_rcv_data_seg *rwqe);

/**
 * @brief Get owner field from CQE function
 *
 * Parse a CQE for its owner field.
 *
 * @param[in] cqe - CQE to parse.
 *
 * @return uint8_t - Owner field value of the CQE.
 */
uint8_t flexio_dev_cqe_get_owner(struct flexio_dev_cqe64 *cqe);

/**
 * @brief Get owner field from EQE function
 *
 * Parse an EQE for its owner field.
 *
 * @param[in] eqe - EQE to parse.
 *
 * @return uint32_t - owner field value of the EQE.
 */
uint8_t flexio_dev_eqe_get_owner(struct flexio_dev_eqe *eqe);

/**
 * @brief Fill out a control send queue wqe segment function
 *
 * Fill the fields of a send WQE segment (4 DWORDs) with control segment information.
 * This should always be the 1st segment of the WQE.
 *
 * @param[in] swqe - Send WQE segment to fill.
 * @param[in] sq_pi - Producer index of the send WQE.
 * @param[in] sq_number - SQ number that holds the WQE.
 * @param[in] ce - wanted CQ policy for CQEs. Value is taken from cq_ce_mode enum.
 * @param[in] ctrl_seg_type - Type of control segment.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_swqe_seg_ctrl_set(union flexio_dev_sqe_seg *swqe, uint32_t sq_pi,
						 uint32_t sq_number, uint32_t ce,
						 flexio_ctrl_seg_t ctrl_seg_type);

/**
 * @brief Fill out an ethernet send queue wqe segment function
 *
 * Fill the fields of a send WQE segment (4 DWORDs) with Ethernet segment information.
 *
 * @param[in] swqe - Send WQE segment to fill.
 * @param[in] cs_swp_flags - Flags for checksum and swap, see PRM section 8.9.4.2,
 *			     Send WQE Construction Summary.
 * @param[in] mss - Maximum Segment Size - For LSO WQEs - the number of bytes in the
 *		    TCP payload to be transmitted in each packet. Must be 0 on non LSO WQEs.
 * @param[in] inline_hdr_bsz - Length of inlined packet headers in bytes.
 *			       This includes the headers in the inline_data segment as well.
 * @param[in] inline_hdrs - First 2 bytes of the inlined packet headers.

 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_swqe_seg_eth_set(union flexio_dev_sqe_seg *swqe,
						uint16_t cs_swp_flags, uint16_t mss,
						uint16_t inline_hdr_bsz, uint8_t inline_hdrs[2]);

/**
 * @brief Fill out an RDMA send queue wqe segment function
 *
 * Fill the fields of a send WQE segment (4 DWORDs) with RDMA segment information.
 *
 * @param[in] swqe - Send WQE segment to fill.
 * @param[in] rkey - Remote memory access key for the RDMA operation.
 * @param[in] data_addr - Address of the data for the RDMA operation.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_swqe_seg_rdma_set(union flexio_dev_sqe_seg *swqe, uint32_t rkey,
						 uint64_t data_addr);

/**
 * @brief Fill out a memory pointer data send queue wqe segment function
 *
 * Fill the fields of a send WQE segment (4 DWORDs) with memory pointer data segment information.
 *
 * @param[in] swqe - Send WQE segment to fill.
 * @param[in] data_sz - Size of the data.
 * @param[in] lkey - Local memory access key for the data operation.
 * @param[in] data_addr - Address of the data for the data operation.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_swqe_seg_mem_ptr_data_set(union flexio_dev_sqe_seg *swqe,
							 uint32_t data_sz, uint32_t lkey,
							 uint64_t data_addr);

/**
 * @brief Fill out an inline data send queue wqe segment function
 *
 * Fill the fields of a send WQE segment (4 DWORDs) with inline data segment information.
 *
 * @param[in] swqe - Send WQE segment to fill.
 * @param[in] data_sz - Size of the data.
 * @param[in] data - Inline data array (3 DWORDs).
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_swqe_seg_inline_data_set(union flexio_dev_sqe_seg *swqe,
							uint32_t data_sz, uint32_t *data);

/**
 * @brief Fill out an Atomic send queue wqe segment function
 *
 * Fill the fields of a send WQE segment (2 DWORDs) with Atomic segment information.
 * This segment can service a compare & swap or fetch & add operation.
 *
 * @param[in] swqe - Send WQE segment to fill.
 * @param[in] swap_or_add_data - The data that will be swapped in or the data that will be added.
 * @param[in] compare_data - The data that will be compared with. Unused in fetch & add operation.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_swqe_seg_atomic_set(union flexio_dev_sqe_seg *swqe,
						   uint64_t swap_or_add_data,
						   uint64_t compare_data);

/**
 * @brief Increment producer index of an RQ by 1 function
 *
 * Mark a WQE for reuse by incrementing the relevant RQ producer index by 1
 *
 * @param[in] rq_dbr - A pointer to the CQ's doorbell record address.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_dbr_rq_inc_pi(uint32_t *rq_dbr);

/**
 * @brief Set consumer index value for a CQ function
 *
 * Writes an updated consumer index number to a CQ's doorbell record
 *
 * @param[in] cq_dbr - A pointer to the CQ's doorbell record address.
 * @param[in] ci - The consumer index value to update.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_dbr_cq_set_ci(uint32_t *cq_dbr, uint32_t ci);

/**
 * @brief Set producer index value for an SQ function
 *
 * Writes an updated producer index number to an SQ's doorbell record
 *
 * @param[in] sq_dbr - A pointer to the SQ's doorbell record address.
 * @param[in] pi - The producer index value to update.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_dbr_sq_set_pi(uint32_t *sq_dbr, uint32_t pi);

/******************************************************************************/

/** @} */

#endif /* _FLEXIO_DEV_QUEUE_ACCESS_H_ */
